/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * paintapi-gtk - gtk implementation of paint api
 */

#include <gdk/gdk.h>

#include "paintapi-gtk.xh"
#include "paintapi.xh"
#include "gtkutil.xh"
#include "mem.xh"
#include "whls.xh"

#include "plot2.xh"

typedef struct {
      paintapi_t api;

      GdkDrawable *drawable;
      my_plot_t *plot;

} paintapi_gtk_t;

static paintapi_gc_t *gc_new(paintapi_t *api_instance) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   GdkGC *gc = gdk_gc_new(gtk->drawable);
   gdk_gc_set_line_attributes(gc, 1, GDK_LINE_SOLID, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
   return (paintapi_gc_t*)gc;
}

static void gc_free(paintapi_t *api_instance, paintapi_gc_t *gc) {
   //paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;
   if(gc) {
      g_object_unref(gc);
   }
}

static void gc_set_dashes(paintapi_t *api_instance, paintapi_gc_t *gc, int dash_offset, char *dash_list, int dash_list_len) {
   //paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   gdk_gc_set_line_attributes((GdkGC*)gc, 1, GDK_LINE_ON_OFF_DASH, GDK_CAP_NOT_LAST, GDK_JOIN_MITER);
   gdk_gc_set_dashes((GdkGC*)gc, dash_offset, dash_list, dash_list_len);
}

static void gc_set_foreground(paintapi_t *api_instance, paintapi_gc_t *gc, paintapi_rgb_t *color) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   paintapi_rgb_t c = *color;

   if(gtk->plot->draw.inverse_colors) {
      WHLSColor whls;
      rgb_to_whls(&c, &whls);
      whls.wv = 1 - whls.wv;
      whls_to_rgb(&whls, &c);
   }

   GdkColor col = {
      .red = c.r,
      .green = c.g,
      .blue = c.b,
   };
   
   GdkColormap *colormap = gdk_colormap_get_system();
   gdk_colormap_alloc_color(colormap, &col, 0, 0);
   gdk_gc_set_foreground((GdkGC*)gc, &col);
}

static void gc_set_function(paintapi_t *api_instance, paintapi_gc_t *gc, paintapi_function_t function) {
   //paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   switch(function) {
      case PAINTAPI_FUNCTION_SET:
	 gdk_gc_set_function((GdkGC*)gc, GDK_SET);
	 break;
      case PAINTAPI_FUNCTION_XOR:
	 gdk_gc_set_function((GdkGC*)gc, GDK_XOR);
	 break;
   }
}

static void draw_line(paintapi_t *api_instance, paintapi_gc_t *gc, int x1, int y1, int x2, int y2) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   gdk_draw_line(gtk->drawable, (GdkGC*)gc, x1, y1, x2, y2);
}

static void draw_lines(paintapi_t *api_instance, paintapi_gc_t *gc, paintapi_point_t *point_list, int point_list_len) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   gdk_draw_lines(gtk->drawable, (GdkGC*)gc, (GdkPoint*)point_list, point_list_len);
}

static void draw_rectangle(paintapi_t *api_instance, paintapi_gc_t *gc, pa_boolean filled, int x1, int y1, int x2, int y2) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   x_gdk_draw_rectangle(gtk->drawable, (GdkGC*)gc, filled, x1, y1, x2, y2);
}

static void draw_text(paintapi_t *api_instance, paintapi_gc_t *gc, void *font, int x1, int y1, const char *text) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   PangoLayout *layout = pango_layout_new(gtk->plot->pangoctx);
   pango_layout_set_text(layout, text, -1);
   pango_layout_set_font_description(layout, gtk->plot->pangofont);
   gdk_draw_layout(gtk->drawable, (GdkGC*)gc, x1, y1, layout);
}

static void draw_textlayout(paintapi_t *api_instance, paintapi_gc_t *gc, int x1, int y1, paintapi_textlayout_t *layout) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   gdk_draw_layout(gtk->drawable, (GdkGC*)gc, x1, y1, (PangoLayout *)layout);
}

static paintapi_textlayout_t *textlayout_create(paintapi_t *api_instance, void *font, const char *text) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   PangoLayout *layout = pango_layout_new(gtk->plot->pangoctx);
   pango_layout_set_text(layout, text, -1);
   pango_layout_set_font_description(layout, gtk->plot->pangofont);
   return (paintapi_textlayout_t*)layout;
}

static void textlayout_free(paintapi_t *api_instance, paintapi_textlayout_t *layout) {
   //paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   // TODO
}

static void textlayout_calculate_size(paintapi_t *api_instance, paintapi_textlayout_t *layout, paintapi_rectangle_t *size) {
   //paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   PangoRectangle ri;
   pango_layout_get_pixel_extents((PangoLayout*)layout, &ri, NULL);

   size->x = ri.x;
   size->y = ri.y;
   size->w = ri.width;
   size->h = ri.height;
}

static void textlayout_set_strikeout(paintapi_t *api_instance, paintapi_textlayout_t *layout) {
   //paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   PangoAttribute *pa_st = pango_attr_strikethrough_new(1);
   pa_st->start_index = 0;
   pa_st->end_index = strlen(pango_layout_get_text((PangoLayout*)layout));
   PangoAttrList *pal = pango_attr_list_new();
   pango_attr_list_insert(pal, pa_st);
   // TODO free pa_st ?
   pango_layout_set_attributes((PangoLayout*)layout, pal);
   // TODO free pal ?
}

static void api_free(paintapi_t *api_instance) {
   paintapi_gtk_t *gtk = (paintapi_gtk_t*)api_instance;

   g_object_unref(gtk->drawable);

   bzero(gtk, sizeof(paintapi_gtk_t));
   free(gtk);
}

paintapi_t *paintapi_gtk_new(GdkDrawable *drawable, my_plot_t *plot) {
   paintapi_gtk_t *gtk = zmalloc_c(sizeof(paintapi_gtk_t));

   gtk->api.api_free = api_free;

   gtk->api.gc_new = gc_new;
   gtk->api.gc_free = gc_free;
   gtk->api.gc_set_dashes = gc_set_dashes;
   gtk->api.gc_set_foreground = gc_set_foreground;
   gtk->api.gc_set_function = gc_set_function;

   gtk->api.draw_line = draw_line;
   gtk->api.draw_lines = draw_lines;
   gtk->api.draw_rectangle = draw_rectangle;
   gtk->api.draw_text = draw_text;
   gtk->api.draw_textlayout = draw_textlayout;

   gtk->api.textlayout_create = textlayout_create;
   gtk->api.textlayout_free = textlayout_free;
   gtk->api.textlayout_calculate_size = textlayout_calculate_size;
   gtk->api.textlayout_set_strikeout = textlayout_set_strikeout;

   gtk->drawable = drawable;
   gtk->plot = plot;

   g_object_ref(drawable);

   return (paintapi_t*)gtk;
}
