/*
 * Plot - plot time-based data on screen and to file with interactive controls
 * Copyright (C) 2006  Jonas Berlin <xkr47@outerspace.dyndns.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****
 * plot2 - most functions related to representing & managing the
 * actual graphs along with gtk event handlers - except for the
 * plotting itself, see plot_draw
 */

#include <ctype.h>
#include <stdio.h>
#include <gtk/gtkfilechooserdialog.h>
#include <gtk/gtkfontsel.h>
#include <gtk/gtktable.h>
#include <gtk/gtkcombobox.h>
#include <gtk/gtkcheckmenuitem.h>
#include <gtk/gtkbutton.h>
#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkversion.h>
#include <gtk/gtkscale.h>
#include <gtk/gtkrange.h>
#include <glib/giochannel.h>
#include <time.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "gnu_extensions.h"
#include "stringutil.xh"
#include "grid.xh"
#include "mem.xh"
#include "util.xh"
#include "gtkutil.xh"
#include "plot2.xh"
#include "export.xh"
#include "plot_draw.xh"
#include "array.xh"
#include "main.xh"
#include "whls.xh"
#include "compat.xh"
#include "paintapi-gtk.xh"

static void plot_need_redraw(my_plot_t *plot);
static void pan_right(my_plot_t *plot);

static void about_clicked(GtkMenuItem *about_menuitem, my_plot_callback_data_t *data) {
   glade_xml_new(data->gladefile, "aboutdialog1", NULL);
}

const char *labelNames[] = {
   "File date:",
   "Start time:",
   "End time:",
   "Covered time:",
   "Time format:",
   "Data columns:",
};

typedef struct {
      GtkLabel *labels[SZ(labelNames)];
} adf_data_t;

static void adf_update(GtkFileChooser *chooser, adf_data_t *data) {
   char *filename = gtk_file_chooser_get_preview_filename(chooser);
   my_dataset_t *dataset = scan_dataset(filename, 10);
   g_free(filename);
   gtk_file_chooser_set_preview_widget_active(chooser, 1);
   if(!dataset) {
      for(int i=0; i<SZ(labelNames); ++i) {
	 gtk_label_set_text(data->labels[i], "?");
      }
      return;
   }

   char buf[40];
   int idx = 0;

   gtk_label_set_text(data->labels[idx++], format_time(dataset->last_modified, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], format_time(dataset->te, 0));
   gtk_label_set_text(data->labels[idx++], format_difftime(dataset->te - dataset->ts, 0));
   gtk_label_set_text(data->labels[idx++], dataset->timetype == TIMETYPE_EPOCH ? "epoch" : "relative");
   snprintf(buf, sizeof(buf), "%d", dataset->item_count);
   gtk_label_set_text(data->labels[idx++], buf);

   unref_dataset(dataset);
}


static void add_data_file_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {

   GladeXML *gladexml = glade_xml_new(data->gladefile, "filechooserdialog_add_file", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "filechooserdialog_add_file");

   GtkFileChooser *chooser = GTK_FILE_CHOOSER(dialog);

   gtk_file_chooser_set_select_multiple(chooser, 1);

   GtkWidget *table = gtk_table_new(5, 2, 0);
   gtk_table_set_col_spacings(GTK_TABLE(table), 2);

   adf_data_t adtd;

   for(unsigned i=0; i<SZ(labelNames); ++i) {
      GtkWidget *labelNameW = gtk_label_new(labelNames[i]);
      GtkWidget *labelW = gtk_label_new("?");

      gtk_misc_set_alignment(GTK_MISC(labelNameW), 1.0, 0.5);
      gtk_misc_set_alignment(GTK_MISC(labelW), 0.0, 0.5);

      adtd.labels[i] = GTK_LABEL(labelW);

      gtk_table_attach(GTK_TABLE(table), labelNameW, 0, 1, i, i+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
      gtk_table_attach(GTK_TABLE(table), labelW,     1, 2, i, i+1, GTK_EXPAND|GTK_FILL, 0, 0, 0);
   }
   gtk_widget_show_all(table);

   gtk_file_chooser_set_preview_widget(chooser, table);
   g_signal_connect(chooser, "update-preview", G_CALLBACK(adf_update), &adtd);

   GtkFileFilter *filter;

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, "Text files");
   gtk_file_filter_add_mime_type(filter, "text/plain");
   gtk_file_chooser_add_filter(chooser, filter);

   filter = gtk_file_filter_new();
   gtk_file_filter_set_name(filter, "All files");
   gtk_file_filter_add_pattern(filter, "*");
   gtk_file_chooser_add_filter(chooser, filter);

   if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
      gtk_widget_hide(dialog);

      GSList* files;
      for(files = gtk_file_chooser_get_filenames(chooser); files; files = files->next) {
	 add_datafile(data->plot, files->data);
      }
      g_slist_free(files);
   }

   gtk_widget_destroy(dialog);
}

void remove_data_file(my_plot_t *plot, int idx) {
   plot_draw_remove_data_file(&plot->draw, idx);

   plot_need_redraw(plot);
   update_minmax(plot);
}

void remove_all_data_files(my_plot_t *plot) {
   plot_draw_remove_all_data_files(&plot->draw);

   plot_need_redraw(plot);
   update_minmax(plot);
}

static void reset_gc(my_plot_t *plot) {
   plot_draw_reset_gc(&plot->draw);

   GdkColor bg;
   get_color(plot, &bg, "#000000");
   gdk_window_set_background(plot->drawable, &bg);
}

static void remove_data_file_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {
   GladeXML *gladexml = glade_xml_new(data->gladefile, "dialog_remove_datafile", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "dialog_remove_datafile");

   GtkWidget *combo = glade_xml_get_widget(gladexml, "combobox_datafiles");

   gtk_combo_box_remove_text(GTK_COMBO_BOX(combo), 0);

   for(int lsi=0; lsi<data->plot->draw.line_set_count; ++lsi) {
      gtk_combo_box_append_text(GTK_COMBO_BOX(combo), data->plot->draw.line_set[lsi].dataset->path);
   }

   if(gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
      gtk_widget_hide(dialog);

      int idx = gtk_combo_box_get_active(GTK_COMBO_BOX(combo));
      if(idx >= 0) {
#if GTK_CHECK_VERSION(2,6,0)
	 printf("removing %d. %s\n", idx, gtk_combo_box_get_active_text(GTK_COMBO_BOX(combo)));
#else
	 printf("removing %d.\n", idx);
#endif
	 remove_data_file(data->plot, idx);
      }
   }

   gtk_widget_destroy(dialog);
}

static void remove_all_data_files_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {
   remove_all_data_files(data->plot);
}

static void refresh_clicked(GtkMenuItem *menuitem, my_plot_t *plot) {
   for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
      update_dataset(plot->draw.line_set[lsi].dataset);
   }

   plot_need_redraw(plot);
   update_minmax(plot);
}

#ifdef USE_FAM

static void FAMError(const char *msg) {
   fprintf(stderr, "%s: %s\n", msg, FamErrlist[FAMErrno]);
}

static gboolean fam_callback(GIOChannel *source, GIOCondition condition, my_plot_t *plot) {
   boolean_t updated = 0;
   while(FAMPending(&plot->fc)) {
      FAMEvent fe;
      //printf("FAM event\n");
      if(FAMNextEvent(&plot->fc, &fe) == -1) {
	 FAMError("fammonitorfile");
	 exit(1);
      }

      if(fe.code != FAMChanged) continue;

      //printf("FAM event changed\n");

      // TODO check only the file that the FAMChanged event indicates

      for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
	 if(!plot->draw.line_set[lsi].dataset->fam_monitored) continue;
	 if(plot->draw.line_set[lsi].dataset->fr.reqnum != fe.fr.reqnum) continue;
	 int oldcount = plot->draw.line_set[lsi].dataset->time_count;
	 if(!update_dataset(plot->draw.line_set[lsi].dataset)) {
	    printf("Problems updating %s\n", plot->draw.line_set[lsi].dataset->path);
	    // TODO maybe ask user whether to start from beginning
	    printf("remove watch for %s\n", plot->draw.line_set[lsi].dataset->path);
	    if(FAMCancelMonitor(&plot->fc, &plot->draw.line_set[lsi].dataset->fr)) {
	       fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
	    }
	    plot->draw.line_set[lsi].dataset->fam_monitored = 0;
	 }
	 if(oldcount < plot->draw.line_set[lsi].dataset->time_count) {
	    updated = 1;
	 }
      }
   }

   if(updated) {
      //printf("FAM updated\n");
      plot_need_redraw(plot);
      update_minmax(plot);
      pan_right(plot);
      // avoid zoompoint gliding implied by the zoompoint-maintaining algorithm in data_scrollbar_controller
      data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0);
   }

   return 1;
}

static void enable_fam(my_plot_t *plot) {
   if(!plot->fam_initialized) {
      if(FAMOpen(&plot->fc)) {
	 FAMError("fammonitorfile");
	 return;
      }

      int fd = FAMCONNECTION_GETFD(&plot->fc);

      GIOChannel *fam_giochannel = g_io_channel_unix_new(fd);
      if(!fam_giochannel) {
	 FAMClose(&plot->fc);
	 return;
      }
      plot->fam_evtsrcid = g_io_add_watch(fam_giochannel, G_IO_IN, (GIOFunc)fam_callback, plot);
      g_io_channel_unref(fam_giochannel);
      plot->fam_initialized = 1;
   }

   gtk_widget_set_sensitive(plot->xscroll_off, 0);
   gtk_widget_set_sensitive(plot->xscroll_off_rel, 0);
   gtk_widget_set_sensitive(plot->xscroll_zoompoint, 0);
   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 0);

   pan_right(plot);

   data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0);

   char path[PATH_MAX];

   for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
      if(plot->draw.line_set[lsi].dataset->fam_monitored) continue;
      realpath(plot->draw.line_set[lsi].dataset->path, path);
      printf("add watch for %s\n", plot->draw.line_set[lsi].dataset->path);
      if(FAMMonitorFile(&plot->fc, path, &plot->draw.line_set[lsi].dataset->fr, plot)) {
	 fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
      }
      plot->draw.line_set[lsi].dataset->fam_monitored = 1;
   }
}

static void disable_fam(my_plot_t *plot) {
   if(!plot->fam_initialized) return;

   gtk_widget_set_sensitive(plot->xscroll_off, 1);
   gtk_widget_set_sensitive(plot->xscroll_off_rel, 1);
   gtk_widget_set_sensitive(plot->xscroll_zoompoint, 1);
   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 1);

   for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
      if(!plot->draw.line_set[lsi].dataset->fam_monitored) continue;
      printf("remove watch for %s\n", plot->draw.line_set[lsi].dataset->path);
      if(FAMCancelMonitor(&plot->fc, &plot->draw.line_set[lsi].dataset->fr)) {
	 fprintf(stderr, "Error cancelling fam monitor for %s: %s\n", plot->draw.line_set[lsi].dataset->path, FamErrlist[FAMErrno]);
      }
      plot->draw.line_set[lsi].dataset->fam_monitored = 0;
   }

   gtk_widget_set_sensitive(plot->menuitem_file_refresh, 1);
}

static void menuitem_file_autorefresh_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean newstate = gtk_check_menu_item_get_active(menuitem);
   gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail), newstate);
   if(newstate) {
      enable_fam(plot);
   } else {
      disable_fam(plot);
   }
}

static void checkbutton_tail_clicked(GtkCheckButton *checkbutton, my_plot_t *plot) {
   gboolean newstate = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(checkbutton));
   gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(plot->menuitem_file_autorefresh), newstate);
   if(newstate) {
      enable_fam(plot);
   } else {
      disable_fam(plot);
   }
}

#endif

void plot_init(my_plot_t *plot) {
   bzero(plot, sizeof(my_plot_t));
   plot_draw_init(&plot->draw);
}

void plot_free(my_plot_t *plot) {
   plot_draw_deinit(&plot->draw);

   data_scrollbar_controller_free(plot->dsc_x);
   data_scrollbar_controller_free(plot->dsc_y);

   if(plot->pangoctx && !plot->da) {
      g_free(plot->pangoctx);
   }

#ifdef USE_FAM
   if(plot->fam_initialized) {
      g_source_remove(plot->fam_evtsrcid);
      FAMClose(&plot->fc);
   }
#endif

   bzero(plot, sizeof(my_plot_t));
   free(plot);
}

my_plot_t *plot_clone(my_plot_t *o) {
   my_plot_t *n = malloc_c(sizeof(my_plot_t));
   plot_init(n);

   n->draw.inverse_colors = o->draw.inverse_colors;
   n->draw.samples_per_pixel = o->draw.samples_per_pixel;
   n->fontsize_at_width = o->fontsize_at_width;
   n->scale_font_by_size = o->scale_font_by_size;
   n->draw.show_legend = o->draw.show_legend;

   n->dsc_x = data_scrollbar_controller_clone(o->dsc_x);
   n->dsc_y = data_scrollbar_controller_clone(o->dsc_y);

   n->draw.xstart = o->draw.xstart;
   n->draw.xend = o->draw.xend;
   n->draw.xmin = o->draw.xmin;
   n->draw.xmax = o->draw.xmax;
   n->draw.ymin = o->draw.ymin;
   n->draw.ymax = o->draw.ymax;

   n->draw.minor_color = o->draw.minor_color;
   n->draw.major_color = o->draw.major_color;
   n->draw.sep_color = o->draw.sep_color;
   n->draw.label_color = o->draw.label_color;
   n->draw.cursor_marker_color = o->draw.cursor_marker_color;
   n->draw.legend_bg_color = o->draw.legend_bg_color;
   n->draw.legend_border_color = o->draw.legend_border_color;
   n->draw.legend_text_color = o->draw.legend_text_color;

   int setcount = o->draw.line_set_count;
   my_plot_line_info_set_arr_ensure(n->draw.line_set, setcount);

   for(int lsi = 0; lsi < setcount; ++lsi) {
      my_plot_line_info_set_t *ns, *os;
      ns = &n->draw.line_set[lsi];
      os = &o->draw.line_set[lsi];

      ns->dataset = os->dataset;
      ref_dataset(os->dataset);

      int linecount = os->line_count;
      my_plot_line_info_arr_ensure(ns->line, linecount);

      for(int li = 0; li < linecount; ++li) {
	 my_plot_line_info_t *nsl, *osl;
	 nsl = &ns->line[li];
	 osl = &os->line[li];

	 nsl->data = osl->data;
	 nsl->color = osl->color;
	 nsl->dataset_idx = osl->dataset_idx;
	 nsl->enabled = osl->enabled;
	 nsl->mode = osl->mode;
      }
   }

   return n;
}

static gint8 cursor_marker_dashes[] = { 2, 4 };

static void plot_need_redraw(my_plot_t *plot) {
   //printf("need_redraw\n");
   if(plot->conc == 0) {
      if(plot->da) gtk_widget_queue_draw(plot->da);
   } else {
      plot->normal_expose++;
   }
}

void plot_size_or_font_changed(my_plot_t *plot) {
   int w,h;
   gdk_drawable_get_size(plot->drawable, &w, &h);

   char *font;
   if(plot->scale_font_by_size) {
      font = malloc(strlen(plot->font) + 40);
      char *space = strrchr(plot->font, ' ');
      double size;
      if(!space) {
	 space = font + strlen(font);
	 size = 12;
      } else {
	 size = atof(space+1);
      }
      memcpy(font, plot->font, (size_t)(space-plot->font));

      // first scale according to screen size
      int cur_screen_width = gdk_screen_get_width(gtk_window_get_screen(plot->parent));
      size *= ((double)cur_screen_width) / plot->fontsize_at_width;

      // then scale according to widget size
      double wh = (w + 4.0 * h / 3.0) / 2.0;

      size *= exp2(wh / cur_screen_width - 1);

      snprintf(font + (space-plot->font), (size_t)40, " %.1lf", round(size * 2) / 2);

   } else {
      font = plot->font;
   }

   //printf("font %s\n", font);

   PangoFontDescription *font_desc = pango_font_description_from_string(font);

   if(plot->scale_font_by_size) {
      free(font);
   }

   if(plot->pangofont) {
      pango_font_description_free(plot->pangofont);
   }

   plot->pangofont = font_desc;

   if(!plot->pangoctx) {
      if(plot->da) {
	 plot->pangoctx = gtk_widget_get_pango_context(plot->da);
      } else {
	 plot->pangoctx = gdk_pango_context_get();
      }
   }

   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   PangoLayout *layout = pango_layout_new(plot->pangoctx);
   pango_layout_set_text(layout, format_time((my_time_t)round(xoff), 1), -1);
   pango_layout_set_font_description(layout, font_desc);

   PangoRectangle ri;
   pango_layout_get_pixel_extents(layout, &ri, NULL);

   ri.height++;

   int font_h = ri.y < 0 ? ri.height - ri.y : ri.height;

   plot->draw.font_height = font_h;
   plot->draw.time_axis_height = font_h * 2.3 + ri.y;
   plot->draw.time_stamp_width = ri.width;

   h -= plot->draw.time_axis_height;

   plot->draw.plot_xe = w - 1;
   plot->draw.plot_ye = h - 1;

   data_scrollbar_controller_set_min_interval(plot->dsc_x, (double)w / 2);
}

static void add_to_zoom_history(my_plot_t *plot) {
   if(plot->draw.line_set_count == 0) {
      // don't save history if no plots
      return;
   }

   int idx = plot->zoom_pos + 1;
   if(idx > plot->zoom_history_count) idx = plot->zoom_history_count;

   my_zoom_view_arr_ensure(plot->zoom_history, idx + 1);
   my_zoom_view_t *z = &plot->zoom_history[idx];

   z->xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
   z->yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);
   z->xlength_log = data_scrollbar_controller_get_page_length_log(plot->dsc_x);
   z->ylength_log = data_scrollbar_controller_get_page_length_log(plot->dsc_y);
   z->xscalepos = data_scrollbar_controller_get_zoompoint(plot->dsc_x);
   z->yscalepos = data_scrollbar_controller_get_zoompoint(plot->dsc_y);

   if(idx > 0) {
      gtk_widget_set_sensitive(plot->button_zoom_prev, 1);
   }
   gtk_widget_set_sensitive(plot->button_zoom_next, 0);

   plot->zoom_pos = idx;
}

static void show_all(my_plot_t *plot) {
   double min, max;

   data_scrollbar_controller_get_range(plot->dsc_x, &min, &max);
   data_scrollbar_controller_set_page_length(plot->dsc_x, max - min);

   data_scrollbar_controller_get_range(plot->dsc_y, &min, &max);
   data_scrollbar_controller_set_page_length(plot->dsc_y, max - min);

   add_to_zoom_history(plot);
}

static void pan_right(my_plot_t *plot) {
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);

   double min, max;
   data_scrollbar_controller_get_range(plot->dsc_x, &min, &max);

   data_scrollbar_controller_set_page_off(plot->dsc_x, max - xlength);
}

static void switch_to_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;
   if(idx == plot->zoom_pos) return;

   plot->zoom_pos = idx;

   gtk_widget_set_sensitive(plot->button_zoom_prev, idx > 0);
   gtk_widget_set_sensitive(plot->button_zoom_next, idx < plot->zoom_history_count-1);

   my_zoom_view_t *z = &plot->zoom_history[idx];

   data_scrollbar_controller_set_page_length_log(plot->dsc_x, z->xlength_log);
   data_scrollbar_controller_set_page_length_log(plot->dsc_y, z->ylength_log);
   data_scrollbar_controller_set_page_off(plot->dsc_x, z->xoff);
   data_scrollbar_controller_set_page_off(plot->dsc_y, z->yoff);
   data_scrollbar_controller_set_zoompoint(plot->dsc_x, z->xscalepos);
   data_scrollbar_controller_set_zoompoint(plot->dsc_y, z->yscalepos);
}

static void show_zoom_history(my_plot_t *plot, int idx) {
   if(idx < 0) return;
   if(idx >= plot->zoom_history_count) return;

   my_zoom_view_t *z = &plot->zoom_history[idx];

   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
   double yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);
   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);

   double z__xoff2 = z->xoff + exp(z->xlength_log);
   double z__yoff2 = z->yoff + exp(z->ylength_log);

   int x = (int)(plot->draw.plot_xe * (z->xoff - xoff) / xlength);
   int y = (int)(plot->draw.plot_ye * (z->yoff - yoff) / ylength);
   int x2 = (int)(plot->draw.plot_xe * (z__xoff2 - xoff) / xlength);
   int y2 = (int)(plot->draw.plot_ye * (z__yoff2 - yoff) / ylength);

   y = plot->draw.plot_ye - y;
   y2 = plot->draw.plot_ye - y2;

   plot->bx = x;
   plot->by = y;
   plot->box = x2;
   plot->boy = y2;

   gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
   x_gdk_draw_rectangle(plot->drawable, plot->draw.cursor_marker_gc, 0, x, y, x2, y2);
}

static void hide_zoom_history(my_plot_t *plot) {
   if(plot->box == COORD_UNDEF) return;

   gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
   x_gdk_draw_rectangle(plot->drawable, plot->draw.cursor_marker_gc, 0, plot->bx, plot->by, plot->box, plot->boy);

   plot->box = COORD_UNDEF;
}

static void button_zoom_prev_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos - 1);
}

static void button_zoom_next_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   switch_to_zoom_history(plot, plot->zoom_pos + 1);
}

static void button_show_all_clicked(GtkWidget *button_or_menuitem, my_plot_t *plot) {
   show_all(plot);
}

static void button_zoom_hover(GtkButton *button, my_plot_t *plot) {
   show_zoom_history(plot, plot->zoom_pos + (GTK_WIDGET(button) == plot->button_zoom_prev ? -1 : 1));
}

static void button_zoom_unhover(GtkButton *button, my_plot_t *plot) {
   hide_zoom_history(plot);
}

static void invert_colors_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean newstate = gtk_check_menu_item_get_active(menuitem);
   if(plot->draw.inverse_colors == newstate) return;

   plot->draw.inverse_colors = newstate;

   reset_gc(plot);

   plot_need_redraw(plot);
}

static void font_response(GtkFontSelectionDialog *dialog, gint response, my_plot_callback_data_t *data) {
   if(response == GTK_RESPONSE_OK || response == GTK_RESPONSE_APPLY) {
      if(data->plot->font) free(data->plot->font);
      data->plot->font = g_strdup(gtk_font_selection_dialog_get_font_name(dialog));
      plot_size_or_font_changed(data->plot);
      plot_need_redraw(data->plot);
   }
   data->plot->fontsize_at_width = gdk_screen_get_width(gtk_window_get_screen(data->plot->parent));

   if(response != GTK_RESPONSE_APPLY) {
      gtk_widget_destroy(GTK_WIDGET(dialog));
   }
}

static void font_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *data) {
   GtkFontSelectionDialog *dialog = GTK_FONT_SELECTION_DIALOG(gtk_font_selection_dialog_new("Select font"));
   gtk_widget_show(dialog->apply_button);
   gtk_font_selection_dialog_set_preview_text(dialog, FONT_TEST_STRING);
   gtk_font_selection_dialog_set_font_name(dialog, data->plot->font);
   g_signal_connect(dialog, "response", G_CALLBACK(font_response), data);
   gtk_widget_show(GTK_WIDGET(dialog));
}

typedef struct {
      GtkRange *range_sample_density;
      GtkLabel *label_render_time;
      my_plot_t *plot;
} my_plot_sample_density_callback_data_t;

static void sample_density_response(GtkDialog *dialog, gint response, my_plot_sample_density_callback_data_t *sdcbd) {
   if(response == GTK_RESPONSE_OK || response == GTK_RESPONSE_APPLY) {
      sdcbd->plot->draw.samples_per_pixel = exp10(gtk_range_get_value(sdcbd->range_sample_density));
      plot_need_redraw(sdcbd->plot);
      gdk_window_process_updates(sdcbd->plot->drawable, 0);

      char tmp[30];
      sprintf(tmp, "<b>%.3lf</b> seconds", sdcbd->plot->draw.last_render_time);
      gtk_label_set_markup(sdcbd->label_render_time, tmp);
   }

   if(response != GTK_RESPONSE_APPLY) {
      bzero(sdcbd, sizeof(my_plot_sample_density_callback_data_t));
      free(sdcbd);
      gtk_widget_destroy(GTK_WIDGET(dialog));
   }
}

static gchar* sample_density_format(GtkScale *scale, gdouble newval, gpointer user_data) {
   if(newval < -0.01) {
      return g_strdup_printf("%.1lf pixels per sample", exp10(-newval));
   } else if(newval <= 0.01) {
      return g_strdup("1 sample per pixel");
   } else {
      return g_strdup_printf("%.1lf samples per pixel", exp10(newval));
   }
}

static void sample_density_clicked(GtkMenuItem *menuitem, my_plot_callback_data_t *cbd) {
   my_plot_sample_density_callback_data_t *sdcbd = zmalloc_c(sizeof(my_plot_sample_density_callback_data_t));
   sdcbd->plot = cbd->plot;

   GladeXML *gladexml = glade_xml_new(cbd->gladefile, "dialog_sample_density", NULL);
   GtkWidget *dialog = glade_xml_get_widget(gladexml, "dialog_sample_density");

   sdcbd->range_sample_density = GTK_RANGE(glade_xml_get_widget(gladexml, "hscale_sample_density"));
   gtk_range_set_value(sdcbd->range_sample_density, log10(cbd->plot->draw.samples_per_pixel));

   sdcbd->label_render_time = GTK_LABEL(glade_xml_get_widget(gladexml, "label_render_time"));
   char tmp[30];
   sprintf(tmp, "<b>%.3lf</b> seconds", sdcbd->plot->draw.last_render_time);
   gtk_label_set_markup(sdcbd->label_render_time, tmp);

   g_signal_connect(sdcbd->range_sample_density, "format-value", G_CALLBACK(sample_density_format), sdcbd);
   g_signal_connect(dialog, "response", G_CALLBACK(sample_density_response), sdcbd);

   gtk_widget_show(GTK_WIDGET(dialog));
}

static void show_scrollbars_and_tools_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   gboolean show = gtk_check_menu_item_get_active(menuitem);
   x_gtk_widget_set_visible(plot->box_xscrolls, show);
   x_gtk_widget_set_visible(plot->box_yscrolls, show);
   x_gtk_widget_set_visible(plot->table_tools, show);
}

static void show_legend_left_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   plot->draw.show_legend = LEGEND_SHOW_LEFT;
   plot_need_redraw(plot);
}
static void show_legend_right_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   plot->draw.show_legend = LEGEND_SHOW_RIGHT;
   plot_need_redraw(plot);
}
static void hide_legend_clicked(GtkCheckMenuItem *menuitem, my_plot_t *plot) {
   plot->draw.show_legend = LEGEND_HIDE;
   plot_need_redraw(plot);
}

static void xor_markers(my_plot_t *plot) {
   if(plot->hline >= 0) {
      gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -plot->vline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->drawable, plot->draw.cursor_marker_gc, 0, plot->hline, plot->draw.plot_xe, plot->hline);
      gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->drawable, plot->draw.cursor_marker_gc, plot->vline, 0, plot->vline, plot->draw.plot_ye);
   }

   for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
      for(int li=0; li<plot->draw.line_set[lsi].line_count; ++li) {
	 if(plot->draw.line_set[lsi].line[li].hline >= 0) {
	    gdk_draw_line(plot->drawable,
			  plot->draw.line_set[lsi].line[li].horiz_marker,
			  0, plot->draw.line_set[lsi].line[li].hline,
			  plot->draw.plot_xe, plot->draw.line_set[lsi].line[li].hline);
	 }
      }
   }
}

static void reset_markers(my_plot_t *plot) {
   plot->hline = -1;
   plot->vline = -1;
   for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
      for(int li=0; li<plot->draw.line_set[lsi].line_count; ++li) {
	 plot->draw.line_set[lsi].line[li].hline = -1;
      }
   }
}

void hide_markers(my_plot_t *plot) {
   xor_markers(plot);
   reset_markers(plot);
}

static void show_markers(my_plot_t *plot, int x, int y) {
   if(!plot->draw.minor_gc) {
      // gui not yet initialized
      return;
   }

   gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -x, cursor_marker_dashes, SZ(cursor_marker_dashes));
   gdk_draw_line(plot->drawable, plot->draw.cursor_marker_gc, 0, y, plot->draw.plot_xe, y);
   gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -y, cursor_marker_dashes, SZ(cursor_marker_dashes));
   gdk_draw_line(plot->drawable, plot->draw.cursor_marker_gc, x, 0, x, plot->draw.plot_ye);
   if(plot->hline >= 0) {
      gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -plot->vline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->drawable, plot->draw.cursor_marker_gc, 0, plot->hline, plot->draw.plot_xe, plot->hline);
      gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
      gdk_draw_line(plot->drawable, plot->draw.cursor_marker_gc, plot->vline, 0, plot->vline, plot->draw.plot_ye);
   }
   plot->vline = x;
   plot->hline = y;

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);

   my_time_t xv = (my_time_t)(x * xlength / plot->draw.plot_xe + xoff);
   for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
      int p = my_time_arr_find(plot->draw.line_set[lsi].dataset->time, xv, 0);
      for(int li=0; li<plot->draw.line_set[lsi].line_count; ++li) {
	 if(!plot->draw.line_set[lsi].line[li].enabled) continue;

	 int dataset_idx = plot->draw.line_set[lsi].line[li].dataset_idx;

	 int dy = !__isnan(plot->draw.line_set[lsi].dataset->item[dataset_idx].data[p]) ?
	    (int)(plot->draw.plot_ye - (plot->draw.line_set[lsi].dataset->item[dataset_idx].data[p] - yoff) / ylength * plot->draw.plot_ye) :
	    -1;

	 if(dy >= 0) gdk_draw_line(plot->drawable, plot->draw.line_set[lsi].line[li].horiz_marker, 0, dy, plot->draw.plot_xe, dy);

	 if(plot->draw.line_set[lsi].line[li].hline >= 0) {
	    gdk_draw_line(plot->drawable,
			  plot->draw.line_set[lsi].line[li].horiz_marker,
			  0, plot->draw.line_set[lsi].line[li].hline,
			  plot->draw.plot_xe, plot->draw.line_set[lsi].line[li].hline);
	 }

	 plot->draw.line_set[lsi].line[li].hline = dy;
      }
   }
}

static void plot_move_to(my_plot_t *plot, int x, int y) {

/*
  data_scrollbar_controller_set_zoompoint(plot->dsc_x, ((double)x) / plot->draw.plot_xe);
  data_scrollbar_controller_set_zoompoint(plot->dsc_y, 1 - ((double)y) / plot->draw.plot_ye);
*/

   int xd = x - plot->bx;
   int yd = plot->by - y;

   xd *= plot->bspeed;
   yd *= plot->bspeed;

   //printf("move %d,%d\n", xd, yd);

   double xmin, xmax;
   double ymin, ymax;

   data_scrollbar_controller_get_range(plot->dsc_x, &xmin, &xmax);
   data_scrollbar_controller_get_range(plot->dsc_y, &ymin, &ymax);

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);

   double nxoff = plot->bxoff - xd * xlength / plot->draw.plot_xe;
   double nyoff = plot->byoff - yd * ylength / plot->draw.plot_ye;

   //printf("nt %lf %lf\n", nxoff, nyoff);

   if(nxoff < xmin) nxoff = xmin; else if(nxoff + xlength > xmax) nxoff = xmax - xlength;
   if(nyoff < ymin) nyoff = ymin; else if(nyoff + ylength > ymax) nyoff = ymax - ylength;

   //printf("n %lf %lf\n", nxoff, nyoff);

   data_scrollbar_controller_set_page_off(plot->dsc_x, nxoff);
   data_scrollbar_controller_set_page_off(plot->dsc_y, nyoff);
}

static int inside_range(int x, int min, int max) {
   return x >= min && x <= max;
}

static int check_if_hit_legend_box(my_plot_t *plot, int x, int y) {
   return plot->draw.show_legend &&
      inside_range(x, plot->draw.legend_x, plot->draw.legend_x2) &&
      inside_range(y, plot->draw.legend_y, plot->draw.legend_y2);
}

static my_plot_line_info_t *check_if_hit_legend_item(my_plot_t *plot, int x, int y) {
   for(int lsi = 0; lsi < plot->draw.line_set_count; ++lsi) {
      my_plot_line_info_t *line = plot->draw.line_set[lsi].line;
      int line_count = plot->draw.line_set[lsi].line_count;

      for(int li=0; li<line_count; ++li) {
	 if(inside_range(x, line[li].legend_x, line[li].legend_x2) &&
	    inside_range(y, line[li].legend_y, line[li].legend_y2)) {
	    return &line[li];
	 }
      }
   }

   return NULL;
}

static gboolean plot_button_press_event(GtkWidget *widget, GdkEventButton *event, my_plot_t *plot) {
   int x = event->x;
   int y = event->y;

#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail));
#else
   const boolean_t tailing = 0;
#endif

   switch((int)event->type) {
/*
  case GDK_2BUTTON_PRESS: {

  printf("dpress %d %d\n", x, y);
  plot_move_to(plot, plot->draw.plot_xe/2, plot->draw.plot_ye/2);
  break;
  }
*/

      case GDK_BUTTON_PRESS:
	 // printf("press %d %d\n", x, y);
	 if(plot->bmode == 0) {

	    plot->bx = x;
	    plot->by = y;

	    hide_markers(plot);

	    if(check_if_hit_legend_box(plot, x, y)) {
	       my_plot_line_info_t *line = check_if_hit_legend_item(plot, x, y);

	       if(line) {
		  switch(event->button) {
		     case 1:
			plot->bstate = line->enabled = !line->enabled;
			plot->bmode = 4;

			plot_need_redraw(plot);
			update_minmax(plot);
			break;

			/*
		     default:
			diff mode, TODO scales don't work and we need a time unit also
			plot->bstate = line->mode == LINE_MODE_NORMAL;
			line->mode = plot->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL;
			plot->bmode = 5;

			plot_need_redraw(plot);
			update_minmax(plot);
			break;
			*/
		  }
	       }
	       return 0;
	    }

	    // update scale pos according to press location

	    int snap_x_threshold = 40; // plot->draw.plot_xe / 10;
	    int snap_y_threshold = 40; // plot->draw.plot_ye / 10;
	    int snap_x = x < snap_x_threshold ? 0 : x >= plot->draw.plot_xe - snap_x_threshold ? plot->draw.plot_xe-1 : x;
	    int snap_y = y < snap_y_threshold ? 0 : y >= plot->draw.plot_ye - snap_y_threshold ? plot->draw.plot_ye-1 : y;

	    data_scrollbar_controller_set_zoompoint(plot->dsc_x, ((double)snap_x) / plot->draw.plot_xe);
	    data_scrollbar_controller_set_zoompoint(plot->dsc_y, 1 - ((double)snap_y) / plot->draw.plot_ye);

	    plot->bspeed = event->state & GDK_SHIFT_MASK ? 4 : 1;

	    if(event->button > 1) {
	       plot->bx = x;
	       plot->bmode = 2;
	       plot->bxloglength = data_scrollbar_controller_get_page_length_log(plot->dsc_x);
	       plot->byloglength = data_scrollbar_controller_get_page_length_log(plot->dsc_y);

	       // when close to border, zoom to border
	       if(tailing || x >= plot->draw.plot_xe - 30) {
		  data_scrollbar_controller_set_zoompoint(plot->dsc_x, 1.0);
	       } else if(x < 30) { // ##TODO## conf 30
		  data_scrollbar_controller_set_zoompoint(plot->dsc_x, 0.0);
	       }

	       if(y < 30) {
		  data_scrollbar_controller_set_zoompoint(plot->dsc_y, 1.0);
	       } else if(y >= plot->draw.plot_ye - 30) {
		  data_scrollbar_controller_set_zoompoint(plot->dsc_y, 0.0);
	       }

	    } else if(event->state & GDK_CONTROL_MASK) {
	       plot->bmode = 3;
	       plot->box = COORD_UNDEF;

	    } else {
	       plot->bx = tailing ? plot->draw.plot_xe : x;
	       plot->bmode = 1;
	       plot->bxoff = data_scrollbar_controller_get_page_off(plot->dsc_x);
	       plot->byoff = data_scrollbar_controller_get_page_off(plot->dsc_y);
	    }
	 }
	 break;
   }

   return 0;
}

#define LIMIT(v,min,max) do{ if(v<(min)) v=min; else if(v>(max)) v=max; } while(0)

static void plot_motion_generic(my_plot_t *plot, int x, int y, int release) {
#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail));
#else
   const boolean_t tailing = 0;
#endif

   switch(plot->bmode) {
      case 0:
	 if(check_if_hit_legend_box(plot, x, y)) {
	    hide_markers(plot);
	 } else {
	    show_markers(plot, x, y);
	 }
	 break;

      case 1:
	 plot_move_to(plot, tailing ? plot->draw.plot_xe : x, y);
	 if(release) {
	    add_to_zoom_history(plot);
	 }
	 break;

      case 2: {
	 // ##TODO## conf 5.0 and 2.0
	 data_scrollbar_controller_set_page_length_log(plot->dsc_x, plot->bxloglength - (x - plot->bx) * plot->bspeed * 5.0 / plot->draw.plot_xe);
	 data_scrollbar_controller_set_page_length_log(plot->dsc_y, plot->byloglength - (y - plot->by) * plot->bspeed * 2.0 / plot->draw.plot_ye);
	 if(release) {
	    add_to_zoom_history(plot);
	 }
	 break;
      }

      case 3: {
	 LIMIT(plot->bx, 0, plot->draw.plot_xe);
	 LIMIT(plot->by, 0, plot->draw.plot_ye);
	 LIMIT(x, 0, plot->draw.plot_xe);
	 LIMIT(y, 0, plot->draw.plot_ye);

	 gdk_gc_set_dashes(plot->draw.cursor_marker_gc, -plot->hline, cursor_marker_dashes, SZ(cursor_marker_dashes));
	 if(!release) {
	    x_gdk_draw_rectangle(plot->drawable, plot->draw.cursor_marker_gc, 0, plot->bx, plot->by, x, y);
	 }
	 if(plot->box != COORD_UNDEF) {
	    x_gdk_draw_rectangle(plot->drawable, plot->draw.cursor_marker_gc, 0, plot->bx, plot->by, plot->box, plot->boy);
	 }
	 if(!release) {
	    plot->box = x;
	    plot->boy = y;
	 } else {
	    int x1 = plot->bx;
	    int y1 = plot->by;

	    y = plot->draw.plot_ye - y;
	    y1 = plot->draw.plot_ye - y1;

	    if(x < x1) { int t = x; x = x1; x1 = t; }
	    if(y < y1) { int t = y; y = y1; y1 = t; }

	    double ox = data_scrollbar_controller_get_page_off(plot->dsc_x);
	    double oy = data_scrollbar_controller_get_page_off(plot->dsc_y);
	    double lx = data_scrollbar_controller_get_page_length(plot->dsc_x);
	    double ly = data_scrollbar_controller_get_page_length(plot->dsc_y);

	    //printf("%d,%d %d,%d\n", x1, y1, x-x1, y-y1);

	    //printf("%lf,%lf %lf,%lf\n", ox, oy, lx, ly);

	    ox += lx * x1 / plot->draw.plot_xe;
	    oy += ly * y1 / plot->draw.plot_ye;

	    lx *= (x - x1) / (double)plot->draw.plot_xe;
	    ly *= (y - y1) / (double)plot->draw.plot_ye;

	    //printf("%lf,%lf %lf,%lf\n\n", ox, oy, lx, ly);

	    // after area zoom, always put center as zoom point
	    data_scrollbar_controller_set_zoompoint(plot->dsc_x, 0.5);
	    data_scrollbar_controller_set_zoompoint(plot->dsc_y, 0.5);

	    // update page lengths
	    data_scrollbar_controller_set_page_length(plot->dsc_x, lx);
	    data_scrollbar_controller_set_page_length(plot->dsc_y, ly);

	    // now, in case either page length subceeded the minimum page length, we adjust offset so that the selected zoom area becomes centered in view
	    double nlx = data_scrollbar_controller_get_page_length(plot->dsc_x);
	    double nly = data_scrollbar_controller_get_page_length(plot->dsc_y);

	    ox -= (nlx - lx) / 2;
	    oy -= (nly - ly) / 2;

	    // update offsets
	    data_scrollbar_controller_set_page_off(plot->dsc_x, ox);
	    data_scrollbar_controller_set_page_off(plot->dsc_y, oy);

	    //printf("%lf,%lf %lf,%lf\n\n", data_scrollbar_controller_get_page_off(plot->dsc_x), data_scrollbar_controller_get_page_off(plot->dsc_y), nlx, nly);

	    plot->box = COORD_UNDEF;

	    add_to_zoom_history(plot);
	 }
	 break;
      }

      case 4: {
	 if(!check_if_hit_legend_box(plot, x, y)) break;

	 my_plot_line_info_t *line = check_if_hit_legend_item(plot, x, y);

	 if(line) {
	    int changed = line->enabled != plot->bstate;
	    if(changed) {
	       line->enabled = plot->bstate;

	       plot_need_redraw(plot);
	       update_minmax(plot);
	    }
	 }
	 break;
      }

      case 5: {
	 if(!check_if_hit_legend_box(plot, x, y)) break;

	 my_plot_line_info_t *line = check_if_hit_legend_item(plot, x, y);

	 if(line) {
	    int changed = line->mode != (plot->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL);
	    if(changed) {
	       line->mode = plot->bstate ? LINE_MODE_DIFF : LINE_MODE_NORMAL;

	       plot_need_redraw(plot);
	       update_minmax(plot);
	    }
	 }
	 break;
      }

      default:
	 fprintf(stderr, "ERROR: bmode = %d\n", plot->bmode);
	 break;
   }
}

static gboolean plot_button_release_event(GtkWidget *widget, GdkEventButton *event, my_plot_t *plot) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(plot, x, y, 1);

   plot->bmode = 0;

   return 0;
}

static gboolean plot_leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, my_plot_t *plot) {
   hide_markers(plot);

   return 0;
}

static gboolean plot_motion_notify_event(GtkWidget *wid, GdkEventMotion *event, my_plot_t *plot) {
   int x = event->x;
   int y = event->y;

   plot_motion_generic(plot, x, y, 0);

   return 0;
}

/*
  my $co_minor = $#colors - 3;
  my $co_major = $#colors - 2;
  my $co_sep = $#colors - 1;
  my $co_label = $#colors - 0;
*/

static void plot_scrollbar_value_changed(GtkRange *not_always_provided, my_plot_t *plot) {

   double xlength = data_scrollbar_controller_get_page_length(plot->dsc_x);
   double xoff = data_scrollbar_controller_get_page_off(plot->dsc_x);

   double ylength = data_scrollbar_controller_get_page_length(plot->dsc_y);
   double yoff = data_scrollbar_controller_get_page_off(plot->dsc_y);

   plot->draw.xmin = xoff;
   plot->draw.xmax = xoff + xlength;

   plot->draw.ymin = yoff;
   plot->draw.ymax = yoff + ylength;

   double xlo, xhi;
   double ylo, yhi;
   data_scrollbar_controller_get_range(plot->dsc_x, &xlo, &xhi);
   data_scrollbar_controller_get_range(plot->dsc_y, &ylo, &yhi);

   plot->draw.xstart = (my_time_t)xlo;
   plot->draw.xend = (my_time_t)xhi;

   gboolean show_all_enable =
      (xhi-xlo - xlength) / (xhi - xlo) > 1e-10 ||
      (yhi-ylo - ylength) / (yhi - ylo) > 1e-10;

   if(plot->button_show_all) {
      gtk_widget_set_sensitive(plot->button_show_all, show_all_enable);
   }
   if(plot->menuitem_show_all) {
      gtk_widget_set_sensitive(plot->menuitem_show_all, show_all_enable);
   }

   plot_need_redraw(plot);
}

static gboolean plot_configure_event(GtkWidget *widget, GdkEventConfigure *do_not_use, my_plot_t *plot) {
   //printf("  --  configure\n");

   plot_size_or_font_changed(plot);

   return 0;
}

static gboolean plot_map_event(GtkWidget *widget, GdkEvent *do_not_use, my_plot_t *plot) {
   //printf("  --  map\n");

   plot->drawable = plot->da->window;

   plot->draw.paintapi = paintapi_gtk_new(plot->drawable, plot);

   // set the background part 1 - create new style for widget
   GtkStyle *style = gtk_widget_get_style(plot->da);
   style = gtk_style_copy(style);
   gtk_widget_set_style(plot->da, style);

   // set the background part 2
   GdkColor bg;
   get_color(plot, &bg, "#000000");
   gdk_window_set_background(plot->drawable, &bg);

   plot_configure_event(widget, NULL, plot);

   return 0;
}

static gboolean plot_expose_event(GtkWidget *widget, GdkEventExpose *event, my_plot_t *plot) {
   //printf("expose\n");

   if(plot->conc > 0) {
      // still not optimal
      plot->normal_expose++;
      //printf(" -> normal_expose++\n");
      return 0;
   }
   int mytrans = ++plot->dsc_x->num_changes + plot->dsc_y->num_changes;
   plot->conc++;
   //    print "->$mytrans plot->conc\n";

   plot->box = COORD_UNDEF;

   plot_draw(&plot->draw);

   xor_markers(plot);

   if(mytrans != plot->dsc_x->num_changes + plot->dsc_y->num_changes || plot->normal_expose) {
      //printf("%d pending exposes, queue draw\n", plot->normal_expose);
      gtk_widget_queue_draw(plot->da);
      plot->normal_expose = 0;
   }
   plot->conc--;

   //printf("/expose\n");

   return 0;
}

// TODO remove this
int get_color(my_plot_t *plot, GdkColor *dst, const char *str) {
   // printf("get_color %s\n", str);
   GdkColormap *colormap = gdk_colormap_get_system();
   if(!gdk_color_parse(plot->draw.inverse_colors ? "white" : "black", dst)) return 1; // it's a hack!
   return gdk_colormap_alloc_color(colormap, dst, 0, 0) ? 0 : 2;
}

// TODO remove this
GdkGC *x_gdk_gc_new_with_fg_color(my_plot_t *plot, const char *color) {
   GdkGC *gc = gdk_gc_new(plot->drawable);
   // gdk_gc_copy(line[li].line_gc, plot->da->style->fg_gc[GTK_STATE_NORMAL]);
   GdkColor c;
   get_color(plot, &c, color);
   gdk_gc_set_foreground(gc, &c);
   return gc;
}

#define VDIFFMIN_DIVISOR 1000

void update_minmax(my_plot_t *plot) {
   my_time_t ts = 0, te = 1;
   double vmin = INFINITY, vmax = -INFINITY;
   double vdiffmin = VDIFFMIN_DIVISOR;

   int count = 0;

   for(int lsi=0; lsi<plot->draw.line_set_count; ++lsi) {
      my_dataset_t *dataset = plot->draw.line_set[lsi].dataset;
      my_plot_line_info_t *line = plot->draw.line_set[lsi].line;
      int line_count = plot->draw.line_set[lsi].line_count;
      for(int li=0; li<line_count; ++li) {
	 if(!line[li].enabled) continue;
	 int d = line[li].dataset_idx;
	 switch(line[li].mode) {
	    case LINE_MODE_NORMAL: {
	       if(count == 0 || vmin > dataset->item[d].min) vmin = dataset->item[d].min;
	       if(count == 0 || vmax < dataset->item[d].max) vmax = dataset->item[d].max;

	       // TODO should we include disabled lines in vdiffmin calculation...
	       double vdiff = dataset->item[d].max - dataset->item[d].min;
	       if(count == 0 || vdiffmin > vdiff) vdiffmin = vdiff;

	       break;
	    }
	    case LINE_MODE_DIFF: {
	       if(count == 0 || vmin > dataset->item[d].min_diff) vmin = dataset->item[d].min_diff;
	       if(count == 0 || vmax < dataset->item[d].max_diff) vmax = dataset->item[d].max_diff;

	       // TODO see above
	       double vdiff = dataset->item[d].max_diff - dataset->item[d].min_diff;
	       if(count == 0 || vdiffmin > vdiff) vdiffmin = vdiff;

	       break;
	    }
	 }
	 count++;
      }

      if(lsi == 0 || dataset->ts < ts) {
	 ts = dataset->ts;
      }

      if(lsi == 0 || dataset->te > te) {
	 te = dataset->te;
      }
   }

   // don't do anything if no lines visible
   if(count == 0) return;

   data_scrollbar_controller_set_range(plot->dsc_x, (double)ts, (double)te);
   //data_scrollbar_controller_set_page_length(plot->dsc_x, (double)(te - ts)); // initially zoom out

   vmin -= (vmax-vmin)/20;
   vmax += (vmax-vmin)/20;

   data_scrollbar_controller_set_range(plot->dsc_y, vmin, vmax);
   //data_scrollbar_controller_set_page_length(plot->dsc_y, vmax-vmin); // initially zoom out

   // TODO maybe use diff data as base for minimum length instead of a fraction of the full range

   double yminlength = vdiffmin > 0 ? vdiffmin / VDIFFMIN_DIVISOR : 1;
   data_scrollbar_controller_set_min_interval(plot->dsc_y, yminlength);

   //add_to_zoom_history(plot);
}

// ##TODO## make local
// ##TODO## the colors suck
paintapi_rgb_t default_line_colors[] = {
   { 0xffff, 0x4444, 0x4444 },
   { 0xffff, 0xc0c0, 0x4040 },
   { 0x4040, 0xffff, 0x4040 },
   { 0x4040, 0xffff, 0xffff },
   { 0x4040, 0x8080, 0xffff },
   { 0xffff, 0x0000, 0x0000 },
   { 0x0000, 0xffff, 0x0000 },
   { 0x0000, 0x0000, 0xffff },
   { 0xffff, 0xffff, 0xffff },
};

int linecoloroff = 0;

static void update_progress(double percent, void *plotp) {
   my_plot_t *plot = (my_plot_t*)plotp;

   gtk_progress_bar_set_fraction(plot->progressbar, percent);
}

void add_datafile(my_plot_t *plot, char *filename) {
   printf("Loading %s ..\n", filename);

   char tmp[40 + strlen(filename)];
   sprintf(tmp, "Loading %s ..", filename);
   gtk_widget_show(GTK_WIDGET(plot->progressbar));
   gtk_progress_bar_set_text(plot->progressbar, tmp);

   gtk_widget_set_sensitive(plot->menuitem_file_add_data_file, 0);

   my_dataset_t *dataset = read_dataset(filename, update_progress, plot);

   gtk_widget_set_sensitive(plot->menuitem_file_add_data_file, 1);

   gtk_progress_bar_set_text(plot->progressbar, "");
   gtk_progress_bar_set_fraction(plot->progressbar, 0.0);
   gtk_widget_hide(GTK_WIDGET(plot->progressbar));

   if(!dataset) {
      if(!quit) {
	 fprintf(stderr, "\033[31;1mWARNING: File '%s' could not be loaded\033[m\n", filename);
      }
      return;
   }

   dataset->maxdiff = 100 * TIME_MULTIPLIER;

   int lsi = plot->draw.line_set_count;

   my_plot_line_info_set_arr_ensure(plot->draw.line_set, lsi + 1);

   plot->draw.line_set[lsi].dataset = dataset;

   printf("Datafile %d: %d data columns, %d samples\n", lsi + 1, dataset->item_count, dataset->time_count);

   for(int dsi=0; dsi<dataset->item_count; ++dsi) {
      int li = plot->draw.line_set[lsi].line_count;
      my_plot_line_info_arr_ensure(plot->draw.line_set[lsi].line, li+1);
      int linecoloridx = linecoloroff++;
      plot->draw.line_set[lsi].line[li].color = default_line_colors[linecoloridx % SZ(default_line_colors)];
      plot->draw.line_set[lsi].line[li].dataset_idx = dsi;
      plot->draw.line_set[lsi].line[li].enabled = 1;
      printf("  column %d: value range %lf - %lf\n", dsi+1, dataset->item[dsi].min, dataset->item[dsi].max);
   }
   puts("");

#ifdef USE_FAM
   boolean_t tailing = gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(plot->checkbutton_tail));
   if(tailing) {
      enable_fam(plot);
   }
#endif

   plot_need_redraw(plot);
   update_minmax(plot);
   show_all(plot);
}

/*
static void screen_size_changed(GdkScreen *screen, my_plot_callback_data_t *data) {
   // works only with xrandr, not with ctrl-alt-+ and ctrl-alt--
   printf("screen %d,%d\n", gdk_screen_get_width(screen), gdk_screen_get_height(screen));
}
*/

typedef enum { DND_INFO_URI, DND_INFO_TEXT } my_drag_drop_info_t;

static void drag_n_drop(GtkWidget *widget, GdkDragContext *context, gint x, gint y, GtkSelectionData *data, guint info, guint drag_time, my_plot_t *plot) {
   switch(info) {
      case DND_INFO_URI: {
	 gchar **uris = gtk_selection_data_get_uris(data);
	 if(!uris) break;
	 for(int i=0; uris[i]; ++i) {
	    printf("Got %s\n", uris[i]);
	    char *hostname;
	    char *path = g_filename_from_uri(uris[i], &hostname, NULL);
	    if(path && (!hostname || !strncmp(hostname, "localhost", (size_t)9))) {
	       add_datafile(plot, path);
	    }
	    if(path) free(path);
	    if(hostname) free(hostname);
	 }
	 g_strfreev (uris);
	 break;
      }
      case DND_INFO_TEXT: {
	 gchar *text = gtk_selection_data_get_text(data);
	 if(!text) break;
	 // assume text is a set of lines, each line containing a filename
	 gchar **paths = g_strsplit_set(text, "\r\n", 0);
	 for(int i=0; paths[i]; ++i) {
	    if(strspn(paths[i], " \t") == strlen(paths[i])) continue; // only whitespace, if any
	    add_datafile(plot, paths[i]);
	 }
	 g_strfreev(paths);
	 free(text);
	 break;
      }
   }
}

void plot_connect_signals(my_plot_t *plot, const char *gladefile, GladeXML *gladexml) {

   my_plot_callback_data_t *cbd = malloc_c(sizeof(my_plot_callback_data_t));
   cbd->plot = plot;
   cbd->gladefile = gladefile;

   // TODO cbd is not freed

   // setup drag and drop

   gtk_drag_dest_set(plot->da, GTK_DEST_DEFAULT_ALL, NULL, 0, GDK_ACTION_COPY);
   GtkTargetList *target_list = gtk_target_list_new(NULL, 0);
   gtk_target_list_add_uri_targets(target_list, DND_INFO_URI);
   gtk_target_list_add_text_targets(target_list, DND_INFO_TEXT);
   gtk_drag_dest_set_target_list(plot->da, target_list);
   gtk_target_list_unref(target_list);
   g_signal_connect(plot->da, "drag-data-received", G_CALLBACK(drag_n_drop), plot);

/*
   GdkScreen* screen = gtk_window_get_screen(plot->parent);
   g_signal_connect(screen, "size-changed", G_CALLBACK(screen_size_changed), cbd);
*/

/*
   GdkRectangle r;
   gdk_screen_get_monitor_geometry(screen, 0, &r);

   printf("screen %d,%d\n", gdk_screen_get_width(screen), gdk_screen_get_height(screen));
   printf("  mm   %d,%d\n", gdk_screen_get_width_mm(screen), gdk_screen_get_height_mm(screen));
   printf("  moni %d,%d\n", r.width, r.height);
*/

   gtk_widget_add_events(plot->da, GDK_LEAVE_NOTIFY_MASK);
   g_signal_connect(plot->da, "leave-notify-event", G_CALLBACK(plot_leave_notify_event), plot);

   gtk_widget_add_events(plot->da, GDK_POINTER_MOTION_MASK);
   g_signal_connect(plot->da, "motion-notify-event", G_CALLBACK(plot_motion_notify_event), plot);

   //gtk_widget_add_events(plot->da, GDK_BUTTON_MOTION_MASK);
   //g_signal_connect(plot->da, "-event", G_CALLBACK(plot_motion_notify_event), plot);

   gtk_widget_add_events(plot->da, GDK_BUTTON_PRESS_MASK);
   g_signal_connect(plot->da, "button-press-event", G_CALLBACK(plot_button_press_event), plot);

   gtk_widget_add_events(plot->da, GDK_BUTTON_RELEASE_MASK);
   g_signal_connect(plot->da, "button-release-event", G_CALLBACK(plot_button_release_event), plot);


   g_signal_connect(plot->da, "map-event", G_CALLBACK(plot_map_event), plot);
   g_signal_connect(plot->da, "expose-event", G_CALLBACK(plot_expose_event), plot);
   g_signal_connect(plot->da, "configure-event", G_CALLBACK(plot_configure_event), plot);

   // buttons

   plot->button_show_all = glade_xml_get_widget(gladexml, "button_show_all");
   g_signal_connect(plot->button_show_all, "clicked", G_CALLBACK(button_show_all_clicked), plot);

   g_signal_connect(plot->button_zoom_prev, "clicked", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(plot->button_zoom_next, "clicked", G_CALLBACK(button_zoom_next_clicked), plot);
   g_signal_connect(plot->button_zoom_prev, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_next, "enter", G_CALLBACK(button_zoom_hover), plot);
   g_signal_connect(plot->button_zoom_prev, "leave", G_CALLBACK(button_zoom_unhover), plot);
   g_signal_connect(plot->button_zoom_next, "leave", G_CALLBACK(button_zoom_unhover), plot);

   //g_signal_connect(glade_xml_get_widget(gladexml, "button_menu"), "clicked", G_CALLBACK(menu_clicked), glade_xml_get_widget(gladexml, "menubar_main"));

   // menu

   plot->box_xscrolls = glade_xml_get_widget(gladexml, "box_xscrolls");
   plot->box_yscrolls = glade_xml_get_widget(gladexml, "box_yscrolls");
   plot->table_tools = glade_xml_get_widget(gladexml, "table_tools");

#ifdef USE_FAM

   plot->menuitem_file_refresh = glade_xml_get_widget(gladexml, "menuitem_file_refresh");
   plot->menuitem_file_autorefresh = glade_xml_get_widget(gladexml, "menuitem_file_autorefresh");
   plot->checkbutton_tail = glade_xml_get_widget(gladexml, "checkbutton_tail");

   plot->xscroll_off = glade_xml_get_widget(gladexml, "xscroll_scrollbar_off");
   plot->xscroll_off_rel = glade_xml_get_widget(gladexml, "xscroll_scrollbar_off_rel");
   plot->xscroll_zoompoint = glade_xml_get_widget(gladexml, "xscroll_scale");

   gtk_widget_set_sensitive(plot->menuitem_file_autorefresh, 1);
   gtk_widget_set_sensitive(plot->checkbutton_tail, 1);

   g_signal_connect(plot->menuitem_file_autorefresh, "activate", G_CALLBACK(menuitem_file_autorefresh_clicked), plot);
   g_signal_connect(plot->checkbutton_tail, "toggled", G_CALLBACK(checkbutton_tail_clicked), plot);

#endif

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_add_data_file"), "activate", G_CALLBACK(add_data_file_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_data_file"), "activate", G_CALLBACK(remove_data_file_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_remove_all_data_files"), "activate", G_CALLBACK(remove_all_data_files_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_refresh"), "activate", G_CALLBACK(refresh_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_file_export_as"), "activate", G_CALLBACK(export_as_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_quit"), "activate", G_CALLBACK(request_quit), (gpointer)2);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_previous"), "activate", G_CALLBACK(button_zoom_prev_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_view_next"), "activate", G_CALLBACK(button_zoom_next_clicked), plot);

   plot->menuitem_show_all = glade_xml_get_widget(gladexml, "menuitem_show_all");
   g_signal_connect(plot->menuitem_show_all, "activate", G_CALLBACK(button_show_all_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_inverse_colors"), "activate", G_CALLBACK(invert_colors_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_font"), "activate", G_CALLBACK(font_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_sample_density"), "activate", G_CALLBACK(sample_density_clicked), cbd);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_legend_left"), "activate", G_CALLBACK(show_legend_left_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_legend_right"), "activate", G_CALLBACK(show_legend_right_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_hide_legend"), "activate", G_CALLBACK(hide_legend_clicked), plot);
   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_show_scrollbars_and_tools"), "activate", G_CALLBACK(show_scrollbars_and_tools_clicked), plot);

   g_signal_connect(glade_xml_get_widget(gladexml, "menuitem_about"), "activate", G_CALLBACK(about_clicked), cbd);

   // scrollbars

   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "xscroll_scrollbar_off"))), "value-changed", G_CALLBACK(plot_scrollbar_value_changed), plot);
   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "xscroll_scrollbar_off"))), "changed", G_CALLBACK(plot_scrollbar_value_changed), plot);
   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "yscroll_scrollbar_off"))), "value-changed", G_CALLBACK(plot_scrollbar_value_changed), plot);
   g_signal_connect(gtk_range_get_adjustment(GTK_RANGE(glade_xml_get_widget(gladexml, "yscroll_scrollbar_off"))), "changed", G_CALLBACK(plot_scrollbar_value_changed), plot);

   plot_scrollbar_value_changed(NULL, plot);
}
